open Definitions

type t = (symbol, int) Hashtbl.t

exception Symbol_not_in_alphabet
exception Multiply_defined_symbol

let empty (): t = Hashtbl.create Common.max_symb

let add (s: symbol) (arity:int) (alpha: t): unit = 
  if Hashtbl.mem alpha s then 
    raise Multiply_defined_symbol
  else
    Hashtbl.add alpha s arity

let arity (s: symbol) (alpha: t): int =
  try
    Hashtbl.find alpha s
  with
    | Not_found -> 
        raise Symbol_not_in_alphabet

let mem s alpha =
  Hashtbl.mem alpha s

let card = Hashtbl2.length 

let to_list = Hashtbl2.elements

let to_string t = 
  Hashtbl.fold
    (fun sym ar s ->
       Symbol.to_string sym ^ ":" ^ (string_of_int ar) ^ s
    )
    t ""


(************************************ TESTS ***********************************)
IFDEF TEST THEN    
open Debug

let _TEST_ALPHABET_ = 
  print_string "Alphabet.ml\n\n";
  let alpha = empty () in
    add "a" 0 alpha;
    add "f" 2 alpha;
    add "b" 0 alpha;
    add "c" 0 alpha;
    asserteq (mem "a" alpha) true "";
    asserteq (mem "g" alpha) false "";
    asserteq (arity "f" alpha) 2 ""
END

    

