open Definitions

type automaton = {
  trans : Transitions.t;
  fstates : Stateset.t;
}


val transtable : automaton -> Transitions.transtable
val copy : automaton -> automaton
val external_to_TA : Stateset.t -> Transitions.external_trans list -> automaton
val to_string : automaton -> string
val to_timbuk : automaton -> string

IFDEF TEST THEN
val to_string_check : automaton -> string
END

type s2t = (state, Term.term) Hashtbl.t
val new_s2t : unit -> ('a, 'b) Hashtbl.t
val s2t_to_string : (int, Term.term) Hashtbl.t -> string


type statevar = (state, Definitions.variable) Hashtbl.t

val new_statevar : unit -> ('a, 'b) Hashtbl.t
val statevar_add : statevar -> state -> Definitions.variable -> unit
val statevar_find : ('a, 'b) Hashtbl.t -> 'a -> 'b
val statevar_to_string : (int, string) Hashtbl.t -> string



val oriented_run :
  Transitions.t -> Term.term -> state -> bool

val all_run :
  Transitions.t -> Term.term -> state list

val normalize_term :
  Transitions.t ->
  Term.term -> bool -> state * Transitions.external_trans list

val normalize_termE :
  Term.term ->
  Transitions.t ->
  state * Formula.t * Transitions.external_trans list

val normalize_new_states :
  Term.term ->
  Transitions.t -> state * Transitions.external_trans list

val oldnormalize_term :
  bool -> Term.term ->
  state -> Transitions.t -> Transitions.external_trans list

val term_to_automaton :
  Term.term list ->
  symbol list ->
  s2t -> automaton * statevar * state list


val not_empty_intersection : (state * state) list -> Transitions.t -> Formula.t

val inter_of' : automaton -> automaton -> Formula.t
val refine : automaton -> automaton -> int

val export : automaton -> automaton 
val export_epsi : automaton -> automaton 

IFDEF TEST THEN
val aut1 : automaton
val aut2 : automaton
val aut10 : automaton
val sv: statevar
END







