(* Implementation des intervals d'entiers *)
open Pervasives2

type interval = { mutable left: int;
                  mutable right: int
                } (* ] left; right [ *)



module Implem = Set2.Make
  (struct
     type t = interval

     let compare = compare (* En cas de probleme sur les max et les min: cmp *)

     let cmp x y = x.left - y.left

     let to_string iv =
       if iv.left + 1 < iv.right then
         Format.sprintf "] %i; %i [" iv.left iv.right
       else
         "∅"
   end)

type t = Implem.t ref

let make l r =
  if l < r then
    ref (Implem.singleton { left = l - 1; right = r + 1 })
  else
    raise (Invalid_argument "Intervals.make")

exception Found of interval

let find p dom =
  try
    Implem.iter (fun iv -> if p iv then raise (Found iv)) !dom;
    raise Not_found
  with Found iv -> iv

let mem dom x =
  try
    ignore (find (fun iv -> iv.left < x && x < iv.right) dom);
    true
  with Not_found -> false


let remove x dom =
  try
    let iv =
      find (fun iv -> iv.left < x && x < iv.right) dom
    in (* x ∈ iv = ] left; right [ *)
      if x - iv.left <= 1 then      (* ] left; x [ = ∅ *)
        if iv.right - x <= 1 then   (* ] x; right [ = ∅ *)
          dom := Implem.remove iv !dom    (* iv = [ x ], we remove it *)
        else
          iv.left <- x
      else
        let right = iv.right in
          iv.right <- x;
          if right - x > 1 then   (* ] x; right [ <> ∅ *)
            dom := Implem.add { left = x; right = right } !dom
  with
    | Not_found -> ()


exception End

let add x dom =
  try
    let borders = ref [] in
      (try
         Implem.iter
           (fun iv ->
              if x = iv.left || x = iv.right then (
                borders := iv :: !borders;
                if List.length !borders = 2 then
                  raise End
              ) else if iv.left < x && x < iv.right then
                raise Break (* x ∈ iv = ] left; right [ *)
           ) !dom
       with End -> ()
      );
      match !borders with
        | [] ->
            dom := Implem.add { left = x - 1; right = x + 1 } !dom
        | [iv] ->
            if iv.left = x then iv.left <- x - 1 else iv.right <- x + 1
        | iv1::iv2::_ ->
            let lborder, rborder =
              if iv1.left = x then iv2, iv1 else iv1, iv2
            in
              rborder.left <- lborder.left;
              dom := Implem.remove lborder !dom
  with Break -> ()


let to_string dom =
  Implem.to_string ~sep:" " ~left:"" ~right:"" !dom


let iter f dom =
  Implem.iter
    (fun iv ->
       for i = iv.left + 1 to iv.right - 1 do
         f i
       done
    ) !dom

let fold f dom =
  Implem.fold
    (fun iv acc ->
       let acc = ref acc in
         for i = iv.left + 1 to iv.right - 1 do
           acc := f i !acc
         done;
         !acc
    ) !dom

let cardinal dom =
  Implem.fold
    (fun iv count ->
       count + (iv.right - iv.left - 1)
    ) !dom 0

let for_all p dom =
  try
    Implem.iter
      (fun iv ->
         for i = iv.left + 1 to iv.right - 1 do
           if not (p i) then raise Break
         done
      ) !dom;
    true
  with Break -> false

let exists p dom =
  try
    Implem.iter
      (fun iv ->
         for i = iv.left + 1 to iv.right - 1 do
           if (p i) then raise Break
         done
      ) !dom;
    false
  with Break -> true

let is_empty dom = Implem.is_empty !dom

let min dom = (Implem.min_elt !dom).left + 1

let max dom = (Implem.max_elt !dom).right - 1

let l_extend dom min =
  try
    let min_iv = Implem.min_elt !dom in
      if min - 1 < min_iv.left then
        min_iv.left <- min - 1
      else
        raise (Invalid_argument "Intervals.l_extend: min")
  with Not_found -> raise (Invalid_argument "Intervals.r_extend: dom")

let r_extend dom max =
  try
    let max_iv = Implem.max_elt !dom in
      if max + 1 > max_iv.right then
        max_iv.right <- max + 1
      else
        raise (Invalid_argument "Intervals.r_extend: max")
  with Not_found -> raise (Invalid_argument "Intervals.r_extend: dom")
