%{
(* ..................................................................
    Module.......: Parser
    Initial Date : 21 may 2007
    Last Update..: 21 may 2007
    Developer....: Yohan Boichut
    Maintenance..: Yohan Boichut and Thomas Genet
    Compiler.....: OCAML 3.09.3
    Comments.....: Parser for Timbuk


  APPROX err20 Identifier ApproxStates err21 RULES ApproxRulesList
  {PApp($3,$4,$7)}
 .................................................................... *)

open Global
open Common

let functionnal_symbols= Global.alpha (* = ref [] *)
let variables = Global.varset (* = ref [] *)
let gstates = ref []

%}

%token OPS VARS TRS APPROX EQUATIONS RULES SET
%token AUTOMATON TRANSITIONS PRIOR STATES FINAL
%token COLON GIVES EQUAL JOIN AND NOT
%token LB RB LP RP COMMA
%token IF IMP EQU
%token PATTERN SUBPATTERN DOTPAT
%token <int> INT
%token <string> STATE
%token <string> IDENT
%token SPEVAR
%token EOF

%start timbukspec
%type < Global.timbukspec > timbukspec

%%

timbukspec :
  constructorsSection
  variablesSection
  trsSection
  automataSection
  patternSection
  dotPatternSection
  approxSection
  EOF                           { ($3, $4 , $5, $6, $7) }
;

dotPatternSection :
| DOTPAT patterns               { $2 }
|                               { [] }
;
patternSection :
| SUBPATTERN patterns           { Common.subpattern := true ; $2 }
| PATTERN patterns              { Common.subpattern := false; $2 }
|                               { [] }
;
patterns :
| term                  {  [$1]  }
| term patterns         { $1::$2 }
;



constructorsSection :
| OPS constructors            { functionnal_symbols := $2; $2 }
|                             { [] }
;
constructors :
| constructor                            {  [$1]  }
| constructor constructors               { $1::$2 }
;
constructor :
| IDENT COLON INT                       { ($1, $3) }
;

variablesSection :
| VARS variables                        { variables := $2; $2 }
|                                       { [] }
;
variables :
| IDENT variables                        { $1::$2 }
| IDENT                                  {  [$1]  }
;


trsSection :
| trs trsSection                        { $1::$2 }
| trs                                   {  [$1]  }
;
trs :
| TRS IDENT trsrules                    { ($2, $3) }
;
trsrules :
| trsrule trsrules                      { $1::$2 }
| trsrule                               {  [$1]  }
;
trsrule :
| term GIVES term IF conditions         { Conditionnal ($1, $3, $5) }
| term GIVES term                       { Usual ($1, $3) }
;
conditions :
| condition                                     {  [$1]  }
| condition AND conditions                      { $1::$3 }
;
condition :
| IDENT JOIN IDENT      { if (List.mem $1 !variables) && (List.mem $3 !variables) then
                            PJoinable (PVar $1, PVar $3)
                          else (
	                    ErrorHandler.errorNumber:= 65;
	                    raise Parsing.Parse_error
                          )
                        }
;



term :
| IDENT                 { try
                            if List.mem $1 !variables then
	                      PVar $1
                            else
	                      let arity = List.assoc $1 !functionnal_symbols in
	                        if arity = 0 then
                                  PFsym ($1, [])
	                        else (
	                          ErrorHandler.errorNumber := 44;
	                          raise Parsing.Parse_error
                                )
                          with
                            | Not_found ->
	                        ErrorHandler.errorNumber := 45;
	                        raise Parsing.Parse_error
                        }
| IDENT LP subterms RP          { try
                                    if List.mem $1 !variables then (
	                              ErrorHandler.errorNumber:=62;
	                              raise Parsing.Parse_error
                                    ) else
	                              let arity = List.assoc $1 !functionnal_symbols in
	                                if List.length $3 = arity then
                                          PFsym ($1, $3)
	                                else (
	                                  ErrorHandler.errorNumber := 44;
	                                  raise Parsing.Parse_error;
	                                )
                                  with
                                    | Not_found -> (* symbol non declare *)
	                              	ErrorHandler.errorNumber := 63;
	                                raise Parsing.Parse_error
	                        }

| SPEVAR                        { PVar "_" }
;

subterms :
| term                          {  [$1]  }
| term COMMA subterms           { $1::$3 }
;


automataSection :
| automaton automataSection     { $1::$2 }
|                               {   []   }
;

automaton :
| AUTOMATON IDENT
    STATES states FINAL STATES fStates
    priorSection
    TRANSITIONS transitions  { if !Global.completeTF then
      	                         PAut ($2, (PState Global.specialState)::$4, $7, $8 ,$10)
                               else
	                         PAut ($2, $4, $7, $8 ,$10)
                             }

| SET IDENT termSets         { PSet ($2, $3) }
;


termSets :
| termSet                       {  [$1]  }
| termSet termSets              { $1::$2 }
;

termSet :
| IDENT                         { try (
                                    if (List.mem $1 !variables) then
	                              begin
	                                ErrorHandler.errorNumber:= 64;
	                                raise Parsing.Parse_error;
	                              end
                                    else
	                              let arity = List.assoc $1 !functionnal_symbols in
	                                if (arity = 0) then
                                          PFsym ($1, [])
	                                else
	                                  begin
	                                    ErrorHandler.errorNumber := 44;
	                                    raise Parsing.Parse_error;
	                                  end
                                  )
                                  with
                                    | Parsing.Parse_error -> raise Parsing.Parse_error
                                    | _ ->
	                                begin
	                                  ErrorHandler.errorNumber := 45;
	                                  raise Parsing.Parse_error;
	                                end
                                }

| SPEVAR                        { failwith "No silent variable in a term set." }


| IDENT LP termsSets RP         { try (
	                            if (List.mem $1 !variables) then
	                              begin
	                                ErrorHandler.errorNumber:=62;
	                                raise Parsing.Parse_error;
	                              end
	                            else
	                              let arity = List.assoc $1 !functionnal_symbols in
	                                if (arity = List.length $3) then
                                          PFsym ($1, $3)
	                                else
	                                  begin
	                                    ErrorHandler.errorNumber := 44;
	                                    raise Parsing.Parse_error;
	                                  end
                                  )
                                  with
	                            | Parsing.Parse_error -> raise Parsing.Parse_error
	                            | _ ->
	                                begin
	                                  ErrorHandler.errorNumber := 63;
	                                  raise Parsing.Parse_error;
	                                end
                                }
;

termsSets :
| termSet                       {  [$1]  }
| termSet COMMA termsSets       { $1::$3 }
;


priorSection :
| PRIOR transitions             { $2 }
|                               { [] }
;

statesList2 :
| STATE                         { if ((String.compare $1 Global.specialState)=0) then
                                    raise Global.SpecialStateDeclaration
                                  else
                                    [ PState $1 ]
                                }

| STATE statesList2             { if ((String.compare $1 Global.specialState)=0) then
                                    raise Global.SpecialStateDeclaration
                                  else
                                    PState $1 :: $2
                                }
;

states :
| statesList2                   { gstates := (PState Global.specialState)::$1; $1 }
;

fStates :
| statesList2                   { if List.for_all (fun x -> List.mem x !gstates) $1 then
                                    $1
                                  else (
	                            ErrorHandler.errorNumber:= 48;
	                            raise Parsing.Parse_error
                                  )
                                }
;

transitions :
| transition transitions        { $1::$2 }
|                               {   []   }
;

transition :
| STATE GIVES STATE             { let lhs = PState $1 and rhs = PState $3 in
                                    if List.mem rhs !gstates && (List.mem lhs !gstates) then
				      if $1 = Global.specialState || $3 = Global.specialState then
					raise Global.SpecialStateInEpsilon
				      else
					(lhs, rhs)
                                    else (
      	                              ErrorHandler.errorNumber := 47;
	                              raise Parsing.Parse_error
	                            )
                                }

| Normalizedterm GIVES STATE    { (* Format.printf "normalizedterm: %s\n@?" $3; *)
                                  let rhs = PState $3 in
                                    if List.mem rhs !gstates && $3 <> Global.specialState then
	                              ($1, rhs)
                                    else
				      if $3 = Global.specialState then
					raise Global.SpecialStateAsRHS
				      else (
					ErrorHandler.errorNumber := 47;
					raise Parsing.Parse_error
				      )
                                }
;

Normalizedterm :
| IDENT                         { try
	                            if List.mem $1 !variables then (
	                              ErrorHandler.errorNumber:= 39;
	                              raise Parsing.Parse_error
	                            ) else
	                              let arity = List.assoc $1 !functionnal_symbols in
	                                if arity = 0 then
	                                  PFsym ($1, [])
	                                else (
		                          ErrorHandler.errorNumber := 44;
		                          raise Parsing.Parse_error
	                                )
                                  with
	                            | Not_found ->
	                                ErrorHandler.errorNumber := 45;
	                                raise Parsing.Parse_error
                                }

| IDENT LP Normalizedterms RP   { (* Format.printf "%s(Normalterms) \n@?" $1; *)
                                  try
	                            if List.mem $1 !variables then (
	                              ErrorHandler.errorNumber := 62;
	                              raise Parsing.Parse_error;
	                            ) else
	                              let arity = List.assoc $1 !functionnal_symbols in
	                                if arity = List.length $3 then
		                          PFsym ($1, $3)
	                                else (
		                          ErrorHandler.errorNumber := 44;
		                          raise Parsing.Parse_error
	                                )
	                          with
	                            | Parsing.Parse_error -> raise Parsing.Parse_error
	                            | _ ->
	                                ErrorHandler.errorNumber := 45;
		                        raise Parsing.Parse_error
                                }
;

Normalizedterms :
| STATE COMMA Normalizedterms   { (* Format.printf "etat lu: %s\n@?" $1; *)
                                  let res =  PState $1 in
                                    if List.mem res !gstates then (
				      if $1 = Global.specialState then
					Global.completeTF := true;
 				      res :: $3
				    ) else (
	                              ErrorHandler.errorNumber := 47;
	                              raise Parsing.Parse_error
	                            )
                                }

| STATE                         { (* Format.printf "dernier etat lu: %s\n@?" $1; *)
                                  let res = PState $1 in
                                    if List.mem res !gstates then (
				      if $1 = Global.specialState then
					Global.completeTF := true;
				      [ res ]
                                    ) else (
	                              ErrorHandler.errorNumber:=47;
	                              raise Parsing.Parse_error
                                    )
                                }
;


approxSection :
| approxElem approxSection      { $1::$2 }
|                               {   []   }
;

approxElem :
| EQUATIONS IDENT RULES equations  { PSEquation ($2, $4) }
;

equations :
| simpleEq equations            { let (lhs, rhs) = $1 in
                                    PEquation (lhs, rhs, []) :: $2
                                }
| otherEq equations             { $1::$2 }
|                               {   []   }
;


simpleEq :
| term EQU term                 { ($1, $3) }
;


simpleEqs :
| simpleEq simpleEqs                            { $1::$2 }
|                                               {   []   }
;

otherEq :
| term EQU term IF conditionEqs                 { PEquation ($1, $3, $5) }
| LB term EQU term RB IMP LB simpleEqs RB       { PExteq ($2, $4, $8) }
| LB term COMMA term RB IMP LB simpleEqs RB     { PAny ($2, $4, $8) }
| LB term RB IMP LB simpleEqs RB                { PUnique ($2 ,$6) }
;



conditionEq :
| IDENT NOT EQU term                            { if List.mem $1 !variables then
                                                    PNotEqual (PVar $1, $4)
                                                  else (
	                                            ErrorHandler.errorNumber := 56;
	                                            raise Parsing.Parse_error
                                                  )
                                                }
;

conditionEqs :
| conditionEq                                   {  [$1]  }
| conditionEq AND conditionEqs                  { $1::$3 }
;

%%


(* trailer-and-end *)
