IFDEF TEST THEN
open Debug
END

open Global
open Definitions


type term =
  | Var of variable
  | Fsym of symbol * term list
  | State of state

type t = term


let is_state : term -> bool = function State _ -> true | _ -> false


(* B: unbox state *)
exception Not_a_state

let get_state (t: term) : state =
  match t with State q -> q | _ -> raise Not_a_state

let rec to_string = function
  | Fsym (symbol, []) -> Symbol.to_string symbol
  | Fsym (symbol, l_subterms) ->
      symbol ^ (List2.to_string ~sep:"," ~left:"(" ~right:")" to_string l_subterms)

  | State t -> State.to_string t
  | Var x -> x

(* B: Le nom n'est vraiment pas bien choisi... *)
let list_to_string (tl: term list) =
  List2.to_string ~sep:"\n" ~left:"" ~right:"\n" to_string tl


let rec height = function
  | Fsym (_, subterms) ->
      1 + (List2.fmax height 0 subterms)
  | State _ -> 1
  | Var _ -> 0


(** The function vars builds the list of variables occuring in the given term.
    without repetitions *)
(* Old: let rec vars : term -> variable list = function | Var x -> [x] |
   Fsym (_, lt) -> List.fold_left (fun vl t -> List2.clean (vars t @ vl))
   [] lt | State _ -> []
*)
let vars : term -> variable list =
  let rec add_vars vars = function
    | Var x -> List2.add x vars (* <- qu'est ce qui passe ici pour "_" *)
    | Fsym (_, terms) -> List.fold_left add_vars vars terms
    | State _ -> vars
  in
    add_vars []

(* and vars_of_tlist (lt:term list):variable list = *)
(* B: la fonction fait le contraire non?!?!? si tout va bien on est lineaire non ???? *)
(* B: une fonction de type Bool c'est pas mieux ? *)
exception Non_linear

let rec non_linearity_checking t =
  let rec aux_non_lin (vars :variable list) : term -> variable list = function
    | Var "_" -> vars
    | Var x ->
	if List.mem x vars then raise Non_linear else x::vars
    | Fsym (_, lt) ->
        List.fold_left aux_non_lin vars lt
    | State _ -> vars
  in
  let _ = aux_non_lin [] t in (* B: C'est pas tres elegant... *)
    ()



(* The following function transforms a parsed term (pTerm defined in
   global.ml) into a term.
*)
(* B: Est-il judicieux d'avoir 2 representations des memes objets ?
   On peut certainement se liberer des pTerm definitivement
*)
let rec pTerm_to_term : pTerm -> term = function
  | PVar x -> Var x
  | PFsym (f, l) ->
      Fsym (f, List.map pTerm_to_term l)
  | PState x -> failwith "to do" (* B: interessant *)



(************************************ TESTS ***********************************)
IFDEF TEST THEN
print_string "Term.ml\n\n"
let s0 = 0
let t1 = Fsym ("f", [ Fsym ("a", []) ])
let t2 = Fsym ("f", [ State s0 ])
let tn = Fsym ("f", [ Fsym ("a", []);
                      Fsym ("g", [ Fsym ("a", []); Fsym ("a", []) ]) ])

let tm = Fsym ("f", [ Fsym ("g", [ Fsym ("a", []); Fsym ("a", []) ]);
                      Fsym ("g", [ Fsym ("a", []); Fsym ("a", []) ]) ])

let tm2= Fsym ("g", [ Fsym ("a", []); Fsym ("a", []) ])
let tx = Fsym ("f", [ Fsym ("a", []); State s0; Var "x"; Var "y" ])
let ty = Fsym ("f", [ Fsym ("a", []); State s0; tn; tm ])
let tq0= Fsym ("f", [ State s0; Fsym ("g", [ State s0; State s0 ]) ])

let tmx= Fsym ("f", [ Fsym ("a", []);
                      Fsym ("g", [ Var "x"; Fsym ("f", [ Var "x"; Var "y" ]) ])
                    ])

open Debug

let _TESTS_TERM_ =
  asserteq (to_string t1) "f(a)" "";
  asserteq (to_string t2) "f(q0)" "";
  asserteq (to_string tn) "f(a,g(a,a))" "";
  asserteq (to_string tm) "f(g(a,a),g(a,a))" "";
  asserteq (height t1) 2 "";
  asserteq (height tx) 2 "";
  asserteq (height (Var("x"))) 0 ""
END
